#include "Evalutor.h"
#include "math.h"

bool __inline IsNumber(char a)
{
	return (a>='0' && a<='9');
}

bool __inline IsLetter(char a)
{
	return (a>='a' && a<='z') || (a>='A' && a<='Z') || (a=='_');
}

int ReadInt(char* str,char** ret)
{
	int n=0;
	while(IsNumber(str[n]))
		n++;
	char* str2=new char[n+1];
	strncpy(str2,(LPCSTR)str,n);
	*ret=&str[n];
	return atol(str2);
}

char* ReadSymbol(char* str,char** ret)
{
	int n=0;
	while(IsNumber(str[n]) || IsLetter(str[n]))
		n++;
	char* str2=new char[n+1];
	str2[n]=0;
	strncpy(str2,(LPCSTR)str,n);
	*ret=&str[n];
	return str2;
}

Evalutor::Evalutor()
{
	err=0;
	str=0;
	linstr=0;
}

void Evalutor::InitiateError(int errNum,LPCSTR errDesc)
{
	err=new ErrInf;
	err->Num=errNum;
	err->Description=errDesc;
}

void Evalutor::Parse()
{
	if(err)
	{
		delete err;
		err=0;
	}
	linstr=Parse_(false);
}

InstructionList* Evalutor::Parse_(BYTE mode)
{
	int num;
	char* sy;
	InstructionList* instr=new InstructionList;
	Instruction ni;
	InstructionList* il;
	while(*str)
	{
		if(IsNumber(*str))
		{
			num=ReadInt(str,&str);
			ZeroMemory(&ni,sizeof(ni));
			ni.Type=1;
			ni.Value.value=num;
			instr->push_back(ni);
		}
		else
		if(IsLetter(*str))
		{
			sy=ReadSymbol(str,&str);
			ZeroMemory(&ni,sizeof(ni));
			if(*str=='(')
			{
				str++;
				ni.Type=5;
				FuncCall* fcall=new FuncCall;
				fcall->name=sy;
				while(1)
				{
					InstructionList* ret;
					ret=Parse_(2);
					if(err)
						return 0;
					fcall->params.push_back(*ret);
					if(*(str-1)==')')
						break;
				}
				ni.Value.fcall=fcall;
			}
			else
			{
				ni.Type=4;
				ni.Value.sym=sy;
			}
			instr->push_back(ni);
		}
		else
			switch(*str)
			{
			case '(':
				str++;
				il=Parse_(1);
				if(err)
					return 0;
				ZeroMemory(&ni,sizeof(ni));
				ni.Type=3;
				ni.Value.instr=il;
				instr->push_back(ni);
				break;
			case ')':
				str++;
				if(mode)
					return instr;
				else
				{
					InitiateError(ERR_UNEXPECTED_END_CHARACTER,"Unexpected end of literal");
					return 0;
				}
				break;
			case ',':
				str++;
				if(mode==2)
					return instr;
				else
				{
					InitiateError(ERR_PARAM_SEPARATOR_WITHOUT_FUNCCALL,"Params separator is out of function call");
					return 0;
				}
				break;
			case '+':
			case '*':
			case '-':
			case '/':
			case '~':
			case '^':
				ZeroMemory(&ni,sizeof(ni));
				ni.Type=2;
				ni.Value.op=*str;
				instr->push_back(ni);
				str++;
				break;
			case '|':
				ZeroMemory(&ni,sizeof(ni));
				ni.Type=2;
				if(*(str+1)=='|')
				{
					ni.Value.op=-1;
					str++;
				}
				else
					ni.Value.op='|';
				instr->push_back(ni);
				str++;
				break;
			case '&':
				ZeroMemory(&ni,sizeof(ni));
				ni.Type=2;
				if(*(str+1)=='&')
				{
					ni.Value.op=-2;
					str++;
				}
				else
					ni.Value.op='&';
				instr->push_back(ni);
				str++;
				break;
			default:
				InitiateError(ERR_INVALID_CHAR,"Invalid character");
				return 0;
			}
	}
	if(!mode)
		return instr;
	else
	{
		InitiateError(ERR_UNEXPECTED_END_OF_STRING,"Unexpected end of expression");
		return 0;
	}
}

int Evalutor::Evalute()
{
	if(err)
	{
		delete err;
		err=0;
	}
	return Evalute_(linstr);
}

int Evalutor::Evalute_(InstructionList* l)
{
	unsigned int i;
	int ret;

	// Func()
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==5)
		{
			int ii;
			FuncCall* fcall=l->at(i).Value.fcall;
			UserDefinedFuncDef* fdef=0;
			for(ii=0;ii<FuncDefs.size();ii++)
				if(!strcmp(fcall->name,FuncDefs[ii].name))
				{
					fdef=&FuncDefs[ii];
					break;
				}
			if(!fdef)
			{
				InitiateError(ERR_UNDEFINED_FUNCTION,"The function is not defined");
				return 0;
			}
			int par_count=fcall->params.size();
			int* params=new int[par_count-1];
			for(ii=0;ii<par_count;ii++)
			{
				params[ii]=Evalute_(&fcall->params[ii]);
				if(err)
					return 0;
			}
			ErrInf* nErr;
			nErr=fdef->addr(params,par_count,&ret);
			if(!nErr)
			{
				InitiateError(ERR_FUNCTION_CALL_FAILED,"The function call is failed");
				return 0;
			}
			if(nErr!=(ErrInf*)1)
			{
				err=nErr;
				return 0;
			}
			l->at(i).Type=1;
			l->at(i).Value.value=ret;
		}
	}

	// ()
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==3)
		{
			ret=Evalute_(l->at(i).Value.instr);
			if(err)
				return 0;
			delete l->at(i).Value.instr;
			l->at(i).Type=1;
			l->at(i).Value.value=ret;
		}
	}

	//  -
	for(i=0;i<l->size();i++)
		if(l->at(i).Type==2 && l->at(i).Value.op=='-')
			if(i==0)
			{
				if(l->size()==1)
				{
					InitiateError(ERR_LONELY_UNARY_OPERATOR,"The unary operator is lonely");
					return 0;
				}
				if(l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_UNARY_MINUS_USAGE,"The unary minus operator can be placed only immidiatly before value");
					return 0;
				}
				l->at(i).Type=1;
				l->at(i).Value.value=-l->at(i+1).Value.value;
				l->erase(l->begin()+i+1);
			}

	//  ~
	for(i=0;i<l->size();i++)
		if(l->at(i).Type==2 && l->at(i).Value.op=='~')
			if(i==0)
			{
				if(l->size()==1)
				{
					InitiateError(ERR_LONELY_UNARY_OPERATOR,"The unary operator is lonely");
					return 0;
				}
				if(l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_UNARY_MINUS_USAGE,"The unary minus operator can be placed only immidiatly before value");
					return 0;
				}
				l->at(i).Type=1;
				l->at(i).Value.value=~l->at(i+1).Value.value;
				l->erase(l->begin()+i+1);
			}

	// ^
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==2)
			if(l->at(i).Value.op=='^')
			{
				if(i==0 || i==l->size())
				{
					InitiateError(ERR_INVALID_OPERATOR_POSITION,"The binary operator there is at the end or begin of expression");
					return 0;
				}
				if(l->at(i-1).Type!=1 || l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"The binary operator invalid operands");
					return 0;
				}
				double x=l->at(i-1).Value.value;
				if(l->at(i).Value.op=='^')
					ret=pow(x,l->at(i+1).Value.value);
				BINARY_OPERATOR_ENDS;
			}
	}

	// *,/
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==2)
			if(l->at(i).Value.op=='*' || l->at(i).Value.op=='/')
			{
				if(i==0 || i==l->size())
				{
					InitiateError(ERR_INVALID_OPERATOR_POSITION,"The binary operator there is at the end or begin of expression");
					return 0;
				}
				if(l->at(i-1).Type!=1 || l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"The binary operator invalid operands");
					return 0;
				}
				if(l->at(i).Value.op=='*')
					ret=(l->at(i-1).Value.value)*(l->at(i+1).Value.value);
				if(l->at(i).Value.op=='/')
				{
					if(l->at(i+1).Value.value==0)
					{
						InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"Divizion by zero");
						return 0;
					}
					ret=(l->at(i-1).Value.value)/(l->at(i+1).Value.value);
				}
				BINARY_OPERATOR_ENDS;
			}
	}

	// +,-
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==2)
			if(l->at(i).Value.op=='+' || l->at(i).Value.op=='-')
			{
				if(i==0 || i==l->size())
				{
					InitiateError(ERR_INVALID_OPERATOR_POSITION,"The binary operator there is at the end or begin of expression");
					return 0;
				}
				if(l->at(i-1).Type!=1 || l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"The binary operator invalid operands");
					return 0;
				}
				if(l->at(i).Value.op=='+')
					ret=(l->at(i-1).Value.value)+(l->at(i+1).Value.value);
				if(l->at(i).Value.op=='-')
					ret=(l->at(i-1).Value.value)-(l->at(i+1).Value.value);
				BINARY_OPERATOR_ENDS;
			}
	}

	// &,|
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==2)
			if(l->at(i).Value.op=='&' || l->at(i).Value.op=='|')
			{
				if(i==0 || i==l->size())
				{
					InitiateError(ERR_INVALID_OPERATOR_POSITION,"The binary operator there is at the end or begin of expression");
					return 0;
				}
				if(l->at(i-1).Type!=1 || l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"The binary operator invalid operands");
					return 0;
				}
				if(l->at(i).Value.op=='&')
					ret=(l->at(i-1).Value.value)&(l->at(i+1).Value.value);
				if(l->at(i).Value.op=='|')
					ret=(l->at(i-1).Value.value)|(l->at(i+1).Value.value);
				BINARY_OPERATOR_ENDS;
			}
	}

	// &&(254),||(255)
	for(i=0;i<l->size();i++)
	{
		BYTE A=l->at(i).Value.op;
		if(l->at(i).Type==2)
			if(l->at(i).Value.op==-2 || l->at(i).Value.op==-1)
			{
				if(i==0 || i==l->size())
				{
					InitiateError(ERR_INVALID_OPERATOR_POSITION,"The binary operator there is at the end or begin of expression");
					return 0;
				}
				if(l->at(i-1).Type!=1 || l->at(i+1).Type!=1)
				{
					InitiateError(ERR_INVALID_OPERATOR_COMBINATION,"The binary operator invalid operands");
					return 0;
				}
				if(l->at(i).Value.op==-2)
					ret=(l->at(i-1).Value.value)&&(l->at(i+1).Value.value);
				if(l->at(i).Value.op==-1)
					ret=(l->at(i-1).Value.value)||(l->at(i+1).Value.value);
				BINARY_OPERATOR_ENDS;
			}
	}

	return l->at(0).Value.value;
}

void Evalutor::ExtractSymbols()
{
	if(err)
	{
		delete err;
		err=0;
	}
	ExtractSymbols_(linstr);
}

void Evalutor::ExtractSymbols_(InstructionList* l)
{
	unsigned int i;
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==3)
			ExtractSymbols_(l->at(i).Value.instr);
		if(l->at(i).Type==5)
			for(int ii=0;ii<l->at(i).Value.fcall->params.size();ii++)
				ExtractSymbols_(&l->at(i).Value.fcall->params[ii]);
		if(l->at(i).Type==4)
		{
			char* name;
			name=l->at(i).Value.sym;
			if(!GetVariable(name))
			{
				Variable v;
				v.name=new char[strlen(name)];
				v.name[strlen(name)]=0;
				strcpy(v.name,name);
				Variables.push_back(v);
			}
		}
	}
}

void Evalutor::ResolveSymbols()
{
	if(err)
	{
		delete err;
		err=0;
	}
	ResolveSymbols_(linstr);
}

void Evalutor::ResolveSymbols_(InstructionList* l)
{
	unsigned int i;
	for(i=0;i<l->size();i++)
	{
		if(l->at(i).Type==3)
		{
			ResolveSymbols_(l->at(i).Value.instr);
			if(err)
				return;
		}
		else
		if(l->at(i).Type==5)
			for(int ii=0;ii<l->at(i).Value.fcall->params.size();ii++)
				{
					ResolveSymbols_(&l->at(i).Value.fcall->params[ii]);
					if(err)
						return;
				}
		else
		if(l->at(i).Type==4)
		{
			Variable* v;
			v=GetVariable(l->at(i).Value.sym);
			if(!v)
			{
				InitiateError(ERR_INVALID_SYMBOL,"Invalid symbol");
				return;
			}
			l->at(i).Type=1;
			l->at(i).Value.value=v->value;
		}
	}
}

Variable* Evalutor::GetVariable(char* name)
{
	unsigned int i;
	for(i=0;i<Variables.size();i++)
		if(!strcmp(Variables[i].name,name))
			return &Variables[i];
	return 0;
}

unsigned int Evalutor::CountVariables()
{
	return Variables.size();
}

Variable* Evalutor::GetVariable(unsigned int index)
{
	return &Variables[index];
}