#pragma once

#include <vector>
#include <windows.h>

struct Instruction;
struct ErrInf;

typedef std::vector<Instruction> InstructionList;
typedef std::vector<InstructionList> ParamList;

typedef ErrInf* UserDefinedCallFunction(int* params, WORD par_count,int*ret);

#define ERR_INVALID_CHAR 0x101
#define ERR_UNEXPECTED_END_OF_STRING 0x102
#define ERR_UNEXPECTED_END_CHARACTER 0x103
#define ERR_PARAM_SEPARATOR_WITHOUT_FUNCCALL 0x104

#define ERR_INVALID_OPERATOR_POSITION 0x201
#define ERR_INVALID_OPERATOR_COMBINATION 0x202
#define ERR_LONELY_UNARY_OPERATOR 0x203
#define ERR_INVALID_UNARY_MINUS_USAGE 0x204

#define ERR_INVALID_SYMBOL 0x301

#define ERR_DIVIZION_BY_ZERO 0x401
#define ERR_INVALID_PARAMS_COUNT 0x402
#define ERR_UNDEFINED_FUNCTION 0x403
#define ERR_FUNCTION_CALL_FAILED 0x404
#define ERR_INVALID_PARAM 0x405

#define BINARY_OPERATOR_ENDS l->at(i-1).Value.value=ret; \
		l->erase(l->begin()+i); \
		l->erase(l->begin()+i); \
		i-=2;					\

struct UserDefinedFuncDef
{
	char* name;
	UserDefinedCallFunction *addr;
};

struct FuncCall
{
	char* name;
	ParamList params;
};

struct Instruction
{
	BYTE Type;
	/*
		0 - unknown
		1 - int value
		2 - operation
		3 - InstructionList
		4 - symbol
		5 - function call
	*/
	union
	{
		long value;
		char op;
		InstructionList* instr;
		char* sym;
		FuncCall* fcall;
	} Value;
};

struct Variable
{
	LPSTR name;
	int value;
};

struct ErrInf
{
	LPCSTR Description;
	int Num;
};


class Evalutor
{
protected:
	void InitiateError(int errNum,LPCSTR errDesc);
	InstructionList* linstr;
	InstructionList* Evalutor::Parse_(BYTE Isnt);
	int Evalute_(InstructionList* l);
	void ExtractSymbols_(InstructionList* l);
	void ResolveSymbols_(InstructionList* l);
	std::vector<Variable> Variables;
public:
	Evalutor();
	ErrInf* err;
	char* str;
	void Parse();
	void ExtractSymbols();
	int Evalute();
	Variable* GetVariable(char* name);
	Variable* GetVariable(unsigned int index);
	unsigned int CountVariables();
	void ResolveSymbols();
	std::vector<UserDefinedFuncDef> FuncDefs;
};